"""
Notice : 神兽保佑 ，测试一次通过
//      
//      ┏┛ ┻━━━━━┛ ┻┓
//      ┃　　　　　　 ┃
//      ┃　　　━　　　┃
//      ┃　┳┛　  ┗┳　┃
//      ┃　　　　　　 ┃
//      ┃　　　┻　　　┃
//      ┃　　　　　　 ┃
//      ┗━┓　　　┏━━━┛
//        ┃　　　┃   Author: somewheve
//        ┃　　　┃   Datetime: 2019/7/10 下午5:11  ---> 无知即是罪恶
//        ┃　　　┗━━━━━━━━━┓
//        ┃　　　　　　　    ┣┓
//        ┃　　　　         ┏┛
//        ┗━┓ ┓ ┏━━━┳ ┓ ┏━┛
//          ┃ ┫ ┫   ┃ ┫ ┫
//          ┗━┻━┛   ┗━┻━┛
//
"""
from copy import copy

from ctpbee.constant import Direction, Offset, EVENT_LOG, ContractData, OrderType, OrderRequest
from ctpbee.data_handle.local_position import PositionHolding
from ctpbee.event_engine import Event

from ctpbee_cta.constant import STOPORDER_PREFIX, StopOrder, StopOrderStatus, EVENT_CTA_STOPORDER
from ctpbee_cta.indicator import round_to


class Handler:
    """  单策略实现  """

    def __init__(self, cta, local_symbol):
        self.cta = cta
        self.local_symbol = local_symbol
        self.event_engine = cta.app.event_engine
        self.orderid_strategy_map = {}
        self.strategy_orderid_map = {}
        self.stop_order_count = 0
        self.stop_orders = {}

    def export_log(self, log):
        event = Event(EVENT_LOG, log)
        self.cta.app.event_engine.put(event)

    def send_order(self, direction: Direction, offset: Offset, price: float, volume: float, stop: bool, lock: bool):
        """ 提供外层访问的API """
        contract = self.cta.app.recorder.get_contract(self.local_symbol)
        if contract is None:
            self.export_log(f"委托失败，找不到合约：{self.local_symbol}")
            return ""
        # Round order price and volume to nearest incremental value
        price = round_to(price, contract.pricetick)
        volume = round_to(volume, contract.min_volume)
        if stop:
            if contract.stop_supported:
                return self.send_server_stop_order(contract, direction, offset, price, volume, lock)
            else:
                return self.send_local_stop_order(direction, offset, price, volume, lock)
        else:
            return self.send_limit_order(contract, direction, offset, price, volume, lock)

    def send_limit_order(
            self,
            contract: ContractData,
            direction: Direction,
            offset: Offset,
            price: float,
            volume: float,
            lock: bool
    ):
        """
        Send a limit order to server.
        """
        return self.send_server_order(
            contract,
            direction,
            offset,
            price,
            volume,
            OrderType.LIMIT,
            lock
        )

    def send_server_stop_order(self, contract: ContractData, direction: Direction, offset: Offset, price: float,
                               volume: float, lock: bool):
        """
        发送停止单到服务器
        如果交易服务器支持停止单才能够进行使用
        """
        return self.send_server_order(
            contract,
            direction,
            offset,
            price,
            volume,
            OrderType.STOP,
            lock
        )

    def send_server_order(
            self,
            contract: ContractData,
            direction: Direction,
            offset: Offset,
            price: float,
            volume: float,
            type: OrderType,
            lock: bool
    ):
        """
        发送新的单子给服务器
        """
        # 创建新的order_req.
        original_req = OrderRequest(
            symbol=contract.symbol,
            exchange=contract.exchange,
            direction=direction,
            offset=offset,
            type=type,
            price=price,
            volume=volume,
        )
        self.export_log(f"发单 合约名称: {contract.symbol}, 方向: {direction}, 操作: {offset}")

        # 通过本地持仓携带的转换请求功能 --> 详情见ctpbee下面的data_handle/local_position.py
        req_list = self.cta.app.recorder.position_manager.convert_order_request(original_req, lock)
        # 发单
        local_orderids = []
        for req in req_list:
            local_orderid = self.cta.app.send_order(
                req)
            local_orderids.append(local_orderid)
            self.cta.app.recorder.position_manager.update_order_request(req, local_orderid)

            # Save relationship between orderid and strategy.
            self.orderid_strategy_map[local_orderid] = self.cta
            self.strategy_orderid_map.setdefault(self.cta.cta_name,set()).add(local_orderid)
        return local_orderids

    def send_local_stop_order(
            self,
            direction: Direction,
            offset: Offset,
            price: float,
            volume: float,
            lock: bool
    ):
        """
        发送本地停止单
        """
        self.stop_order_count += 1
        stop_orderid = f"{STOPORDER_PREFIX}.{self.stop_order_count}"

        stop_order = StopOrder(
            local_symbol=self.local_symbol,
            direction=direction,
            offset=offset,
            price=price,
            volume=volume,
            stop_orderid=stop_orderid,
            strategy_name=self.cta.cta_name,
            lock=lock
        )
        self.stop_orders[stop_orderid] = stop_order

        # local_orderids = self.strategy_orderid_map[self.cta.cta_name]
        # local_orderids.add(stop_orderid)
        self.strategy_orderid_map.setdefault(self.cta.cta_name, set()).add(stop_orderid)
        self.put_stop_order_event(stop_order)
        return stop_orderid

    def put_stop_order_event(self, stop_order: StopOrder):
        """
        推送事件到停止单处理函数更新
        """
        event = Event(EVENT_CTA_STOPORDER, stop_order)
        self.event_engine.put(event)

    def cancel_server_order(self, local_orderid: str):
        """
        通过local_orderid来撤掉服务器的停止单
        """
        order = self.cta.app.recorder.get_order(local_orderid)
        if not order:
            self.export_log(f"撤单失败，找不到委托{local_orderid}, {self.cta.cta_name}", )
            return
        # 快速创建OrderReq
        req = order._create_cancel_request()
        self.cta.app.cancel_order(req)

    def cancel_local_stop_order(self, stop_orderid: str):
        """
        撤掉本地的停止单
        """
        stop_order = self.stop_orders.get(stop_orderid, None)
        if not stop_order:
            return

        # 移除关系
        self.stop_orders.pop(stop_orderid)

        local_orderids = self.strategy_orderid_map[self.cta.cta_name]
        if stop_orderid in local_orderids:
            local_orderids.remove(stop_orderid)

        # 更新停止单的状态
        stop_order.status = StopOrderStatus.CANCELLED
        # 推送到处理函数
        self.put_stop_order_event(stop_order)

    def cancel_order(self, local_orderid: str):
        """
        撤单
        """
        if local_orderid.startswith(STOPORDER_PREFIX):
            self.cancel_local_stop_order(local_orderid)
        else:
            self.cancel_server_order(local_orderid)

    def cancel_all(self):
        """
        撤掉全部单子
        """
        local_orderids = self.strategy_orderid_map[self.cta.cta_name]
        if not local_orderids:
            return

        for local_orderid in copy(local_orderids):
            self.cancel_order(local_orderid)



"""
Notice : 神兽保佑 ，测试一次通过
//      
//      ┏┛ ┻━━━━━┛ ┻┓
//      ┃　　　　　　 ┃
//      ┃　　　━　　　┃
//      ┃　┳┛　  ┗┳　┃
//      ┃　　　　　　 ┃
//      ┃　　　┻　　　┃
//      ┃　　　　　　 ┃
//      ┗━┓　　　┏━━━┛
//        ┃　　　┃   Author: somewheve
//        ┃　　　┃   Datetime: 2019/7/10 下午5:09  ---> 无知即是罪恶
//        ┃　　　┗━━━━━━━━━┓
//        ┃　　　　　　　    ┣┓
//        ┃　　　　         ┏┛
//        ┗━┓ ┓ ┏━━━┳ ┓ ┏━┛
//          ┃ ┫ ┫   ┃ ┫ ┫
//          ┗━┻━┛   ┗━┻━┛
//
"""
from ctpbee import CtpBee
from ctpbee.constant import Direction, Offset, EVENT_BAR
from ctpbee.constant import EVENT_LOG
from ctpbee.event_engine import Event

from ctpbee_cta.handler import Handler


class CtaCore:
    def __init__(self, cta_name, app: CtpBee, symbol):
        self.app = app
        self.cta_name = cta_name

        # 核心处理器 --> 调度发单等等
        self.core_handler = Handler(cta=self, local_symbol=symbol)

    @property
    def pos(self):
        """ 返回持仓数量 """
        return len(self.app.recorder.positions)

    def push_bar(self, bar):
        """ 推送bar """
        event = Event(EVENT_BAR, bar)
        self.app.event_engine.put(event)

    def export_log(self, log):
        """ 输出日志信息 """
        event = Event(EVENT_LOG, log)
        self.app.event_engine.put(event)

    def buy(self, price: float, volume: float, stop: bool = False, lock: bool = False):
        """
        开仓 多头
        """
        return self.core_handler.send_order(Direction.LONG, Offset.OPEN, price, volume, stop, lock)

    def sell(self, price: float, volume: float, stop: bool = False, lock: bool = False):
        """ 平空头 """
        return self.core_handler.send_order(Direction.SHORT, Offset.CLOSE, price, volume, stop, lock)

    def short(self, price: float, volume: float, stop: bool = False, lock: bool = False):
        """
         开仓 空头
        """
        return self.core_handler.send_order(Direction.SHORT, Offset.OPEN, price, volume, stop, lock)

    def cover(self, price: float, volume: float, stop: bool = False, lock: bool = False):
        """
        平 多头
        """
        return self.core_handler.send_order(Direction.LONG, Offset.CLOSE, price, volume, stop, lock)



"""
Notice : 神兽保佑 ，测试一次通过
//      
//      ┏┛ ┻━━━━━┛ ┻┓
//      ┃　　　　　　 ┃
//      ┃　　　━　　　┃
//      ┃　┳┛　  ┗┳　┃
//      ┃　　　　　　 ┃
//      ┃　　　┻　　　┃
//      ┃　　　　　　 ┃
//      ┗━┓　　　┏━━━┛
//        ┃　　　┃   Author: somewheve
//        ┃　　　┃   Datetime: 2019/7/10 下午4:58  ---> 无知即是罪恶
//        ┃　　　┗━━━━━━━━━┓
//        ┃　　　　　　　    ┣┓
//        ┃　　　　         ┏┛
//        ┗━┓ ┓ ┏━━━┳ ┓ ┏━┛
//          ┃ ┫ ┫   ┃ ┫ ┫
//          ┗━┻━┛   ┗━┻━┛
//
"""

from dataclasses import dataclass, field
from enum import Enum

from ctpbee.constant import Direction, Offset

APP_NAME = "CtaStrategy"
STOPORDER_PREFIX = "STOP"


class StopOrderStatus(Enum):
    WAITING = "等待中"
    CANCELLED = "已撤销"
    TRIGGERED = "已触发"


class EngineType(Enum):
    LIVE = "实盘"
    BACKTESTING = "回测"


class BacktestingMode(Enum):
    BAR = 1
    TICK = 2


@dataclass
class StopOrder:
    local_symbol: str
    direction: Direction
    offset: Offset
    price: float
    volume: float
    stop_orderid: str
    strategy_name: str
    lock: bool = False
    local_orderids: list = field(default_factory=list)
    status: StopOrderStatus = StopOrderStatus.WAITING


EVENT_CTA_LOG = "eCtaLog"
EVENT_CTA_STRATEGY = "eCtaStrategy"
EVENT_CTA_STOPORDER = "eCtaStopOrder"



"""
Notice : 神兽保佑 ，测试一次通过
//      
//      ┏┛ ┻━━━━━┛ ┻┓
//      ┃　　　　　　 ┃
//      ┃　　　━　　　┃
//      ┃　┳┛　  ┗┳　┃
//      ┃　　　　　　 ┃
//      ┃　　　┻　　　┃
//      ┃　　　　　　 ┃
//      ┗━┓　　　┏━━━┛
//        ┃　　　┃   Author: somewheve
//        ┃　　　┃   Datetime: 2019/7/7 上午11:07  ---> 无知即是罪恶
//        ┃　　　┗━━━━━━━━━┓
//        ┃　　　　　　　    ┣┓
//        ┃　　　　         ┏┛
//        ┗━┓ ┓ ┏━━━┳ ┓ ┏━┛
//          ┃ ┫ ┫   ┃ ┫ ┫
//          ┗━┻━┛   ┗━┻━┛
//
"""

import platform
from setuptools import Extension, setup

ext_modules = []

pkgs = ['ctpbee_cta']
install_requires = ['ctpbee', "dataclasses"]
setup(
    name='ctpbee_cta',
    version='0.11',
    description="ctpbee cta strategy support",
    author='somewheve',
    author_email='somewheve@gmail.com',
    url='https://github.com/ctpbee/ctpbee_cta',
    license="MIT",
    packages=pkgs,
    install_requires=install_requires,
    platforms=["Windows", "Linux", "Mac OS-X"],
    ext_modules=ext_modules,
)



"""
Notice : 神兽保佑 ，测试一次通过
//      
//      ┏┛ ┻━━━━━┛ ┻┓
//      ┃　　　　　　 ┃
//      ┃　　　━　　　┃
//      ┃　┳┛　  ┗┳　┃
//      ┃　　　　　　 ┃
//      ┃　　　┻　　　┃
//      ┃　　　　　　 ┃
//      ┗━┓　　　┏━━━┛
//        ┃　　　┃   Author: somewheve
//        ┃　　　┃   Datetime: 2019/7/7 下午12:21  ---> 无知即是罪恶
//        ┃　　　┗━━━━━━━━━┓
//        ┃　　　　　　　    ┣┓
//        ┃　　　　         ┏┛
//        ┗━┓ ┓ ┏━━━┳ ┓ ┏━┛
//          ┃ ┫ ┫   ┃ ┫ ┫
//          ┗━┻━┛   ┗━┻━┛
//
"""

### 一些有趣的指标 欢迎提供


import numpy as np
import talib

def round_to(value: float, target: float) -> float:
    """
    Round price to price tick value.
    """

    return int(round(value / target)) * target


class ArrayManager(object):
    """
    For:
    1. time series container of bar data
    2. calculating technical indicator value
    """

    def __init__(self, size=100):
        """Constructor"""
        self.count = 0
        self.size = size
        self.inited = False

        self.open_array = np.zeros(size)
        self.high_array = np.zeros(size)
        self.low_array = np.zeros(size)
        self.close_array = np.zeros(size)
        self.volume_array = np.zeros(size)

    def update_bar(self, bar):
        """
        Update new bar data into array manager.
        """
        self.count += 1
        if not self.inited and self.count >= self.size:
            self.inited = True

        self.open_array[:-1] = self.open_array[1:]
        self.high_array[:-1] = self.high_array[1:]
        self.low_array[:-1] = self.low_array[1:]
        self.close_array[:-1] = self.close_array[1:]
        self.volume_array[:-1] = self.volume_array[1:]

        self.open_array[-1] = bar.open_price
        self.high_array[-1] = bar.high_price
        self.low_array[-1] = bar.low_price
        self.close_array[-1] = bar.close_price
        self.volume_array[-1] = bar.volume

    @property
    def open(self):
        """
        Get open price time series.
        """
        return self.open_array

    @property
    def high(self):
        """
        Get high price time series.
        """
        return self.high_array

    @property
    def low(self):
        """
        Get low price time series.
        """
        return self.low_array

    @property
    def close(self):
        """
        Get close price time series.
        """
        return self.close_array

    @property
    def volume(self):
        """
        Get trading volume time series.
        """
        return self.volume_array

    def sma(self, n, array=False):
        """
        Simple moving average.
        """
        result = talib.SMA(self.close, n)
        if array:
            return result
        return result[-1]

    def std(self, n, array=False):
        """
        Standard deviation
        """
        result = talib.STDDEV(self.close, n)
        if array:
            return result
        return result[-1]

    def cci(self, n, array=False):
        """
        Commodity Channel Index (CCI).
        """
        result = talib.CCI(self.high, self.low, self.close, n)
        if array:
            return result
        return result[-1]

    def atr(self, n, array=False):
        """
        Average True Range (ATR).
        """
        result = talib.ATR(self.high, self.low, self.close, n)
        if array:
            return result
        return result[-1]

    def rsi(self, n, array=False):
        """
        Relative Strenght Index (RSI).
        """
        result = talib.RSI(self.close, n)
        if array:
            return result
        return result[-1]

    def macd(self, fast_period, slow_period, signal_period, array=False):
        """
        MACD.
        """
        macd, signal, hist = talib.MACD(
            self.close, fast_period, slow_period, signal_period
        )
        if array:
            return macd, signal, hist
        return macd[-1], signal[-1], hist[-1]

    def adx(self, n, array=False):
        """
        ADX.
        """
        result = talib.ADX(self.high, self.low, self.close, n)
        if array:
            return result
        return result[-1]

    def boll(self, n, dev, array=False):
        """
        Bollinger Channel.
        """
        mid = self.sma(n, array)
        std = self.std(n, array)

        up = mid + std * dev
        down = mid - std * dev

        return up, down

    def keltner(self, n, dev, array=False):
        """
        Keltner Channel.
        """
        mid = self.sma(n, array)
        atr = self.atr(n, array)

        up = mid + atr * dev
        down = mid - atr * dev

        return up, down

    def donchian(self, n, array=False):
        """
        Donchian Channel.
        """
        up = talib.MAX(self.high, n)
        down = talib.MIN(self.low, n)

        if array:
            return up, down
        return up[-1], down[-1]

